# Standard Library
import re
import os
import random
from multiprocessing.dummy import Pool as ThreadPool

# Third Party
import requests
from yaml import load, Loader
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import WebDriverWait

# Local Modules
from git_pull.exceptions import DeniedRequest

PATH_TO_RESOURCES = os.path.join(os.path.dirname(__file__), "resources")

CHROME_OPTIONS = Options()
# CHROME_OPTIONS.add_argument("--disable-extensions")
# CHROME_OPTIONS.add_argument("--disable-gpu")
# CHROME_OPTIONS.add_argument("--no-sandbox") # Linux only
CHROME_OPTIONS.add_argument("--headless")

LANGUAGES = load(
    open(os.path.join(PATH_TO_RESOURCES, "languages.yml"), "r"),
    Loader=Loader
).items()
VENDOR_PATTERNS = load(
    open(os.path.join(PATH_TO_RESOURCES, "vendor.yml"), "r"),
    Loader=Loader
)
DOC_PATTERNS = load(
    open(os.path.join(PATH_TO_RESOURCES, "documentation.yml"), "r"),
    Loader=Loader
)["Files"]
USER_AGENTS = list(load(
    open(os.path.join(PATH_TO_RESOURCES, "useragents.yml"), "r"),
    Loader=Loader
))


def get_parse_tree(url):
    response = requests.get(url, headers={"User-Agent": random.choice(USER_AGENTS)})
    tree = BeautifulSoup(response.text, "lxml")

    # Checks if Github denied the request
    if any(h1.get_text() == "Whoa there!" for h1 in tree.find_all("h1")):
        raise DeniedRequest("Github denied your request for:", url)

    return tree


def get_dynamic_content(driver, xpath, delay=10):
    """
    Helper method for scraping dynamic content from a Selenium webdriver. It
    waits for a particular piece of content to load before scraping, and
    returns an empty list if it waits too long.

    Parameters
    ----------
    driver : ...
        Selenium webdriver instance with a webpage loaded on it
    xpath : ...
        XPATH specifying where to find the dynamic content
    delay : int
        how long (in seconds) to wait for the content to load before aborting

    Returns
    -------
    list of Selenium elements
    """

    try:
        return WebDriverWait(driver, delay).until(lambda d: d.find_elements_by_xpath(xpath))
    except TimeoutException:
        return []


def safe_scrape(func):
    """
    Decorator that wraps scraper functions in a try/except.
    """

    def wrapper(self):
        try:
            return func(self)
        except:
            return None

    return wrapper


# @safe_scrape
def scrape_personal_info(parse_tree, tag, id, callback=None):
    target = parse_tree.find(tag, id)
    if not callback:
        return target.text if target else "N/A"

    return callback(target)


def concurrent_exec(func, iterable, num_threads):
    """
    Executes a function on a list of inputs in parallel by dividing the work
    among multiple threads.

    Parameters
    ----------
    func : ...
        function to execute
    iterable : list
        list of arguments to map the function to
    num_threads : int
        number of threads to allocate

    Returns
    -------
    iterator of function outputs corresponding to each input
    """

    pool = ThreadPool(num_threads)
    map_of_items = pool.map(func, iterable)
    pool.close()
    pool.join()

    return map_of_items


def get_next_page_url(tree):
    buttons = tree.find_all("a", "btn btn-outline BtnGroup-item")
    for button in buttons:
        if button.get_text() == "Next":
            return button["href"]

    return ""


def fetch_repo_names(owner):
    def process_repo_list(tree):
        links = tree.find_all("div", "d-inline-block mb-1")
        for link in links:
            yield link.a.get_text().replace(' ', '').strip()

    # Loads the first page of repos
    url = f"https://github.com/{owner}?page=1&tab=repositories"
    tree = get_parse_tree(url)
    yield from process_repo_list(tree)

    if tree.find_all("div", "paginate-container"):
        while url := get_next_page_url(tree):
            tree = get_parse_tree(url)
            yield from process_repo_list(tree)


def identify_file_type(file_path):
    for language, attrs in LANGUAGES:
        if file_path.lower().endswith(tuple(attrs["extensions"])):
            return language

    for doc_pattern in DOC_PATTERNS:
        if re.search(doc_pattern, file_path):
            return "Documentation"

    return ""


def fetch_file_paths(repo_name, owner):
    # TODO: Find a way to force dynamic content to load without using Selenium

    # driver = webdriver.Remote(SERVICE.service_url, OPTIONS) # Linux only
    driver = webdriver.Chrome(
        executable_path=os.path.join(PATH_TO_RESOURCES, "chromedriver"),
        chrome_options=CHROME_OPTIONS
    ) # OS X only
    driver.get(f"https://github.com/{owner}/{repo_name}/find/master")

    files = get_dynamic_content(
        driver,
        '//span[@class="d-inline-block  js-tree-browser-result-path"]'
    )
    file_paths = []
    for path in files:
        path = path.text
        is_code = False

        for language, attrs in LANGUAGES:
            if path.lower().endswith(tuple(attrs["extensions"])):
                is_vendor_code = False

                for vendor_pattern in VENDOR_PATTERNS:
                    if re.search(vendor_pattern, path):
                        is_vendor_code = True
                        break

                if not is_vendor_code:
                    is_code = True
                    file_paths.append(path)
                    break

        if not is_code:
            for doc_pattern in DOC_PATTERNS:
                if re.search(doc_pattern, path):
                    file_paths.append(path)
                    break

    driver.close()

    return file_paths



class DeniedRequest(Exception):
    """
    Thrown when Github denies a web request.
    """

    pass


class InvalidUsernameError(Exception):
    """
    Thrown when Github username doesn't exist.
    """

    pass



# Standard Library
import urllib.parse
from collections import defaultdict
from datetime import datetime
from functools import partial
from multiprocessing import cpu_count

# Local Modules
import git_pull.utilities as utils
from git_pull.exceptions import InvalidUsernameError


class File(object):
    def __init__(self, path, repo, owner, scrape_everything=False):
        self.path = path
        self.type = utils.identify_file_type(path)
        self.repo, self.owner = repo, owner
        self.raw_url = f"https://raw.githubusercontent.com/{owner}/{repo}/master/{urllib.parse.quote(path.encode('utf-8'))}"
        self.blames = {}

        if scrape_everything:
            self.blames = self.scrape_blames()

    def scrape_blames(self):
        url = f"https://github.com/{self.owner}/{self.repo}/blame/master/{self.path}"
        tree = utils.get_parse_tree(url)

        author_to_line_nums = defaultdict(lambda: {"line_nums": set(), "committers": set()})
        for section in tree.find_all("div", "blame-hunk d-flex border-gray-light border-bottom"): # Commit blocks
            line_nums = [int(ln.get_text()) for ln in section.find_all("div", "blob-num blame-blob-num bg-gray-light js-line-number")]
            author_label = section.find("div", "AvatarStack-body")["aria-label"]

            split_author_label = author_label.split(" and ")
            if len(split_author_label) > 1: # Duet commit (pair programmers)
                author = split_author_label[0]
                committer = split_author_label[1].split(" (non-author committer)")[0]

                author_to_line_nums[author]["committers"].add(committer)
                author_to_line_nums[author]["line_nums"] |= set(line_nums)
            else: # Solo commit
                author_to_line_nums[author_label]["line_nums"] |= set(line_nums)

        self.blames = dict(author_to_line_nums)
        return self.blames

    def to_dict(self):
        return {
            "path": self.path,
            "type": self.type,
            # "owner": self.owner,
            # "repo": self.repo,
            "raw_url": self.raw_url,
            "blames": self.blames
        }


class Repo(object):
    def __init__(self, name, owner, num_threads=cpu_count(), scrape_everything=False):
        self.name, self.owner = name, owner
        self._threads = num_threads
        self._tree = utils.get_parse_tree(f"https://github.com/{owner}/{name}")

        self.files = set()
        self.topics = None
        self.star_count = None
        self.fork_count = None
        self.fork_status = None

        if scrape_everything:
            self.scrape_files(scrape_everything)
            self.scrape_topics()
            self.scrape_star_count()
            self.scrape_fork_count()
            self.scrape_fork_status()

    def scrape_files(self, scrape_everything=False):
        if self._threads > 0:
            files = utils.concurrent_exec(
                partial(self.scrape_file, scrape_everything=scrape_everything),
                utils.fetch_file_paths(self.name, self.owner),
                self._threads
            )
            self.files |= set(files)
        else:
            for file_path in utils.fetch_file_paths(self.name, self.owner):
                file = self.scrape_file(file_path, scrape_everything)
                self.files.add(file)

        return self.files

    def scrape_file(self, file_path, scrape_everything=False):
        return File(
            file_path,
            self.name,
            self.owner,
            scrape_everything
        )

    # @utils.safe_scrape
    # def scrape_maintenance_data(self):
    #     base_url = f"https://github.com/{self.owner}/{self.name}"
    #
    #     def get_count_container(extension, to_remove):
    #         tree = get_parse_tree(f"{base_url}/{extension})
    #         count = tree.find('a', class_="btn-link selected")
    #
    #         if not count:
    #             return 0
    #
    #         return int(count.get_text().strip().replace(to_remove, '').replace(',', ''))
    #
    #     self.open_issues = get_count_container("issues?q=is%3Aopen+is%3Aissue", " Open")
    #     self.closed_issues = get_count_container("issues?q=is%3Aissue+is%3Aclosed", " Closed")
    #
    #     self.open_pulls = get_count_container("pulls?q=is%3Aopen+is%3Apr", " Open")
    #     self.closed_pulls = get_count_container("pulls?q=is%3Apr+is%3Aclosed", " Closed")

    @utils.safe_scrape
    def scrape_topics(self):
        containers = self._tree.find_all('a', "topic-tag topic-tag-link")
        self.topics = [tag.get_text().replace(' ', '').strip() for tag in containers]

        return self.topics

    @utils.safe_scrape
    def scrape_star_count(self):
        star_container = self._tree.find('a', "social-count js-social-count")
        self.star_count = int(star_container["aria-label"].split(" ")[0])

        return self.star_count

    @utils.safe_scrape
    def scrape_fork_count(self):
        fork_container = self._tree.find_all('a', "social-count")[2] # [-1]?
        self.fork_count = int(fork_container.get_text().replace(',', ''))

        return self.fork_count

    @utils.safe_scrape
    def scrape_fork_status(self):
        fork_icon = self._tree.find("svg", "octicon octicon-repo-forked text-gray mr-2")
        self.fork_status = bool(fork_icon)

        return self.fork_status

    def to_dict(self):
        return {
            "name": self.name,
            "owner": self.owner,
            "files": [file.to_dict() for file in self.files],
            "topics": self.topics,
            "star_count": self.star_count,
            "fork_count": self.fork_count,
            "fork_status": self.fork_status
            # "open_issues": self.open_issues,
            # "closed_issues": self.closed_issues,
            # "open_pulls": self.open_pulls,
            # "closed_pulls": self.closed_pulls
        }


class GithubProfile(object):
    def __init__(self, username, num_threads=cpu_count(), scrape_everything=False):
        self.username = username
        self._threads = num_threads
        self._tree = utils.get_parse_tree(f"https://github.com/{username}")

        if not self._tree.find_all("div", "js-yearly-contributions"):
            raise InvalidUsernameError(username)

        self.name = ''
        self.avatar = ''
        self.follower_count = ''
        self.contribution_graph = dict()
        self.location = ''
        self.personal_site = ''
        self.workplace = ''
        self.repos = set()

        if scrape_everything:
            self.scrape_name()
            self.scrape_avatar()
            self.scrape_follower_count()
            self.scrape_contribution_graph()
            self.scrape_location()
            self.scrape_personal_site()
            self.scrape_workplace()
            self.scrape_repos(scrape_everything)

    ## Social Info ##

    def scrape_name(self):
        self.name = utils.scrape_personal_info(
            parse_tree=self._tree,
            tag="span",
            id="p-name vcard-fullname d-block overflow-hidden"
        )
        return self.name

    def scrape_avatar(self):
        self.avatar_url = utils.scrape_personal_info(
            parse_tree=self._tree,
            tag="img",
            id="avatar avatar-user width-full border bg-white",
            callback=lambda t: t["src"]
        )
        return self.avatar_url

    def scrape_follower_count(self):
        link_element = self._tree.find("a", "link-gray no-underline no-wrap")
        self.follower_count = int(link_element.find("span", "text-bold text-gray-dark").get_text())
        return self.follower_count

    def scrape_contribution_graph(self):
        get_map = lambda parse_tree: {
            tile["data-date"]: int(tile["data-count"])
        for tile in parse_tree.find_all("rect", "day")}

        graph = get_map(self._tree)
        for link in self._tree.find_all('a', "js-year-link filter-item px-3 mb-2 py-2"):
            url = f"https://github.com{link['href']}"
            commit_overview = utils.get_parse_tree(url)
            graph = {**graph, **get_map(commit_overview)}

        self.contribution_graph = dict(sorted(
            graph.items(),
            key=lambda c: datetime.strptime(c[0], "%Y-%m-%d")
        ))
        return self.contribution_graph

    def scrape_location(self):
        self.location = utils.scrape_personal_info(self._tree, "span", "p-label")
        return self.location

    def scrape_personal_site(self):
        list_element = self._tree.find("li", attrs={"data-test-selector": "profile-website-url"})
        if list_element:
            link_element = list_element.find("a")
            self.personal_site = link_element["href"]

        return self.personal_site

    def scrape_workplace(self):
        self.workplace = utils.scrape_personal_info(self._tree, "span", "p-org")
        return self.workplace

    ## Repos ##

    def scrape_repos(self, scrape_everything=False):
        for repo_name in utils.fetch_repo_names(self.username):
            repo = GithubProfile.scrape_repo(repo_name, scrape_everything)
            self.repos.add(repo)

        return self.repos

    def scrape_repo(self, repo_name, scrape_everything=False):
        return Repo(repo_name, self.username, self._threads, scrape_everything)

    ## Misc. ##

    def to_dict(self):
        return {
            "name": self.name,
            "avatar_url": self.avatar_url,
            "follower_count": self.follower_count,
            "contribution_graph": self.contribution_graph,
            "location": self.location,
            "personal_site": self.personal_site,
            "workplace": self.workplace,
            "repos": [repo.to_dict() for repo in self.repos]
        }



import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from codecs import open

if sys.version_info[:3] < (3, 0, 0):
    print("Requires Python 3 to run.")
    sys.exit(1)

with open("README.md", encoding="utf-8") as file:
    readme = file.read()

setup(
    name="git-pull",
    description="Parallelized web scraper for Github",
    long_description=readme,
    long_description_content_type="text/markdown",
    version="v1.0.4",
    packages=["git_pull"],
    include_package_data=True,
    python_requires=">=3",
    url="https://github.com/shobrook/git-pull",
    author="shobrook",
    author_email="shobrookj@gmail.com",
    # classifiers=[],
    install_requires=[],
    keywords=["scraper", "web-scraper", "github", "github-scraper", "github-api", "parallel"],
    license="MIT"
)
